// IMPLEMENTATION FILE FOR MATHS.H

#ifndef _MATHS_H
#define _MATHS_H

#pragma warning (disable:4305)
#pragma warning (disable:4244)

#include <windows.h>		// Windows header file
#include <GL\gl.h>			// OpenGL header file
#include <GL\glu.h>			// The OpenGL Glu header file
#include <cmath>
#include <stdio.h>
#include <time.h>

// Defines for maths
#define RANDOM_FLOAT (((float)rand()-(float)rand())/RAND_MAX)
#define ABS(x) (x < 0 ? -(x) : (x))
#define MIN(x,y) (x < y ? x : y)
#define MAX(x,y) (x > y ? x : y)
#define SQUARE(number) (number*number)
#define PI			(3.14159265359f)
#define EPSILON 1e-12
#define DEG2RAD(a)	(PI/180*(a))
#define RAD2DEG(a)	(180/PI*(a))
#define FOV(a,b) RAD2DEG(2.0*atan2(a/2,b))
#define SAFEDEL(x) if(x !=NULL) {delete[] x; x = NULL;}

typedef float VERTEX2fv[2];
typedef float VERTEX3fv[3];
typedef float VERTEX4fv[4];
typedef float MATRIXfv[16];


double FastSin0 (double fAngle);
double FastSin1 (double fAngle);
double FastCos0 (double fAngle);
double FastCos1 (double fAngle);
double FastTan0 (double fAngle);
double FastTan1 (double fAngle);
double FastInvSin (double fValue);
double FastInvCos (double fValue);
double FastInvTan0 (double fValue);
double FastInvTan1 (double fValue);

int RANDOM(int min, int max);

enum RenderMode
{
	RENDER_MODE_SIMPLE = 0,
	RENDER_MODE_DLIST,
	RENDER_MODE_VARRAY
};

class CMatrix;
class CVector;
class CQuaternion;

/* --------------------------------- VECTOR CLASS ----------------------------------------//
//----------------------------------------------------------------------------------------*/

class CVector
{

public:
	union
	{
		struct
		{
			float x;
			float y;
			float z;
			float w;	// HOMOGENEOUS COORDINATE
		};
	float m_pos[4];
	};
		

	// Constructors
	CVector(float x=0, float y=0, float z=0, float w=1) : x(x), y(y), z(z), w(w) {  }
	CVector(const CVector &V) : x(V.x), y(V.y), z(V.z), w(V.w) { }
	const CVector& operator = (const CVector& V);
	const CVector& operator = (const float *d);
	//	Add & Substract, Positive & Negative
	const CVector operator + (const CVector& Vector) const;
	const CVector operator + () const;
	const CVector& operator += (const CVector& Vector);
	
	// minus vector 
	const CVector operator - (const CVector& Vector) const;
	const CVector operator - () const;
	const CVector &operator -= (const CVector& Vector);
	
	//	Multiply & Divide
	const CVector operator * (const CVector& Vector) const;
	const CVector& operator *= (const CVector& Vector);
	const CVector operator * (const float d) const;
	const CVector& operator *= (const float d);
	const CVector operator / (const CVector& Vector) const;
	const CVector &operator /= (const CVector& Vector);
	const CVector operator / (const float d) const;
	const CVector& operator /= (const float d);
	
	const float operator[]( int index ) const { return m_pos[index]; }

	const bool operator > (const CVector& Vector);
	const bool operator < (const CVector& Vector);
	//	Dot Product
	const float Dot (const CVector& Vector) const;
	//	Cross Product
	const CVector Cross(const CVector& Vector) const;
	
	void Transform(CMatrix *matrix);
	void Transform3(CMatrix *matrix);

	const float Magnitude() const;
	
	//	Set length (Normalize if 1) or Magnitude of vector
	void Normalize();
		
	const void SetNormal(CVector vTriangle0, CVector vTriangle1, CVector vTriangle2 );
	
	//	The angle between two vectors in radians
	const float Angle(const CVector& Normal) const;
	
	//	Reflect in Normal Vector
	const CVector Reflection(const CVector& PlaneNormal) const;

	//	Rotate dAngle Degrees (radians) Around Normal
	const CVector Rotate(const float dAngle, const CVector& Normal) const;
	
	// Plane Distance
	float PlaneDistance(CVector Normal, CVector Point);
	
	// Is Plane intersected?
	bool IntersectedPlane(CVector vTriangle[], CVector vLine[]);


};

/* --------------------------------- END VECTOR CLASS ------------------------------------//
//----------------------------------------------------------------------------------------*/




/* -------------------------------- START QUATERNION VECTOR CLASS ------------------------//
//----------------------------------------------------------------------------------------*/

class CQuaternion
{

public:

	union
	{
		struct
		{
			float x, y, z, w;
		};
		float m_quat[4];
	};

	void CreateAxis(float x, float y, float z, float degree);		
	float GetMagnitude();
	void Normalize();
	void EulerToQuat(float roll, float pitch, float yaw);
	CVector QuatToEuler(CQuaternion *euler);
		
	void Rotate(float angle, float x, float y, float z);
	void Derotate(float *angle, CVector axis);

	CQuaternion QuatToAxisAngle();
	CQuaternion AxisAngleToQuat(CVector *axis, float angle);
	CQuaternion SlerpQuat(CQuaternion quat, float slerp);
		
	CQuaternion operator *(CQuaternion quat);
	const CQuaternion& operator = (CQuaternion q)
	{
		x = q.x; y= q.y; z = q.z; w = q.w; 
	}

	const float operator [] (int i)
	{
		return m_quat[i];
	};

	//Dtor / Ctor
	CQuaternion(float x1 = 0.0, float y1 = 0.0, float z1 = 0.0, float w1 = 0.0)  
	{ x = x1; y = y1; z = z1; w = w1;};
		
	~CQuaternion(){};
};
		
	
/* -------------------------------- END QUATERNION VECTOR CLASS --------------------------//
//----------------------------------------------------------------------------------------*/





/* --------------------------------- MATRIX CLASS ----------------------------------------//
//----------------------------------------------------------------------------------------*/

class CMatrix
{

public:
	float m_matrix[16];
			
	CMatrix(){Identity();};
		
	__inline CMatrix& operator = (CMatrix& Matrix)
	{
		for (int i=0; i<16; i++)
			m_matrix[i] = Matrix.m_matrix[i];

		return *this;
	};
	float __inline operator[](int i) {	return m_matrix[i];  };

	void Multiply(CMatrix *matrix);
	void SetTranslation(CVector pos);
	void SetInvTranslation(CVector pos);
	
	void __inline Scale(float x, float y, float z)
	{
		m_matrix[0] = x;   
		m_matrix[5] = y;
		m_matrix[10] = z;  
	};
	void __inline Identity()
	{
		for (int i=0; i<16; i++)
			m_matrix[i] = 0.0;

		m_matrix[0] = m_matrix[5] = m_matrix[10] = m_matrix[15] = 1;
	};
	void RotateDegrees(double dAngleX, double dAngleY, double dAngleZ);
	void RotateRadians(double dAngleX, double dAngleY, double dAngleZ);
	void RotateInvDegrees(double dAngleX, double dAngleY, double dAngleZ);
	void RotateInvRadians(double dAngleX, double dAngleY, double dAngleZ);

	void __inline InverseTranslateVect(float *pVect)
	{
		pVect[0] = pVect[0]-m_matrix[12];
		pVect[1] = pVect[1]-m_matrix[13];
		pVect[2] = pVect[2]-m_matrix[14];
	};

	void __inline InverseRotateVect(float *pVect)
	{
		float vec[3];

		vec[0] = pVect[0]*m_matrix[0]+pVect[1]*m_matrix[1]+pVect[2]*m_matrix[2];
		vec[1] = pVect[0]*m_matrix[4]+pVect[1]*m_matrix[5]+pVect[2]*m_matrix[6];
		vec[2] = pVect[0]*m_matrix[8]+pVect[1]*m_matrix[9]+pVect[2]*m_matrix[10];

		for (int i=0; i<3; i++)
			pVect[i] = vec[i];
	};
};

/* ---------------------------------END MATRIX CLASS -------------------------------------//
//----------------------------------------------------------------------------------------*/



typedef VERTEX3fv Tri_FaceFF[3];

struct tVERTEX2f
{
	tVERTEX2f(float x = 0, float y = 0) : x(x), y(y){};
	float x;
	float y;
};

struct tVERTEX3f
{
	tVERTEX3f(float x=0, float y=0, float z=0) : x(x), y(y), z(z) {};
	tVERTEX3f(CVector v): x(v.x), y(v.y), z(v.z) {};
	float x;
	float y;
	float z;
} ;

struct tVERTEX4f
{
	tVERTEX4f(float x=0, float y=0, float z=0, float w=0) : x(x), y(y), z(z), w(w) {};
	float x;
	float y;
	float z;
	float w;
};

struct tAnchor
{
	bool m_bSelected;
	tVERTEX3f m_Pt;
};

struct tBoneVertex
{
	char m_boneID;
	VERTEX3fv m_Pos;
	
	const tBoneVertex operator = (const CVector& v)
	{
		m_Pos[0] = v.x; m_Pos[1] = v.y; m_Pos[2] = v.z;
		return *this;
	};

	tBoneVertex() 
	{
		m_Pos[0] = 0; m_Pos[2] = 0; m_Pos[1] = 0;
	};
};

struct tMesh
{
	int m_materialIndex;
	int m_numTriangles;
	int *m_pTriangleIndices;

	tMesh(): m_materialIndex(-1), m_numTriangles(-1), m_pTriangleIndices(NULL) {};
};

struct tTriangle
{
	VERTEX3fv m_vertexNormals[3];
	VERTEX3fv m_s, m_t;
	float m_vertexIndices[3];
};

//	Animation keyframe information
struct tKeyframe
{
	int m_jointIndex;
	float m_time;	// in milliseconds
	VERTEX3fv m_parameter;
};

		//	Skeleton bone joint
struct tJoint
{
	CVector m_localRotation;
	CVector m_localTranslation;
	CMatrix m_absolute, m_relative;

	int m_numRotationKeyframes, m_numTranslationKeyframes;
	tKeyframe *m_pTranslationKeyframes;
	tKeyframe *m_pRotationKeyframes;

	int m_currentTranslationKeyframe, m_currentRotationKeyframe;
	CMatrix m_final;

	int m_parent;
};

CVector GetPlaneEq(float *arr, CVector *p1, CVector *p2, CVector *p3);
CVector GetPlaneEq(float *arr, float *p1, float *p2, float *p3);

CVector GetMousePos(int x, int y, int winy, int z, CVector *ret = NULL);

unsigned int FACT(unsigned int x);
int CombC(int N, int x);
void DrawArbitraryAxes(float axissize);

void QuaternionToMatrix(CQuaternion *quat, CMatrix *pMatrix);
void MatrixToQuaternion(CQuaternion *quat, CMatrix *pMatrix);

float PlaneDistance(CVector Normal, CVector Point);
bool IntersectedPlane(CVector vTriangle[], CVector vLine[]);


// RANDOM NUMBER GENERATOR
int Random(int from, int to);
int Random();
void seed();


// VARIOUS USEFUL FUNCTIONS
float DistancePointLine( CVector Point, CVector LineStart, CVector LineEnd);
void DrawSphere(CVector c, double r, int n, unsigned int render);

bool SaveMatrix(CMatrix *tmp, char *filename, FILE *file = NULL);
bool LoadMatrix(CMatrix *tmp, char *filename, FILE *file = NULL);

#endif // MATHS_H
